﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace archtype
{
    class Program
    {
        static void Main(string[] args)
        {
            // Give them the help screen
            if (args.Length == 0 || args[0] == "/?" || args[0] == "-?" || args[0] == "--?" || args[0] == "?" || args[0] == "help" || args[0] == "/help" || args[0] == "--help")
            {
                
                string helpstring = @"
Usage: archtype.exe [/?] [/v] filename

Options:
    -?          Get this help screen.
    -v          Get version information.

Archtype opens up the specified files (or filenames, wildcards * ? are allowed) searches for
the headers of the executable.  It will attempt to identify:

Architecture : Either i386 (32bit), x86_64 (64bit) or IA64(Itanium)
Build Date   : The date the executable was created
Symbols      : Whether or not the executable contains debugging symbols
Flags        : These are the 'Characteristics' flags, reproduced below

Characteristics flags:
(reproduced from: http://msdn.microsoft.com/en-us/library/windows/desktop/ms680313(v=vs.85).aspx )            
IMAGE_FILE_RELOCS_STRIPPED         = 0x0001,   // Relocation information was stripped from the file. The file
                                               // must be loaded at its preferred base address. If the base address
                                               // is not available, the loader reports an error.
IMAGE_FILE_EXECUTABLE_IMAGE        = 0x0002,   // The file is executable (there are no unresolved external references).
IMAGE_FILE_LINE_NUMS_STRIPPED      = 0x0004,   // COFF line numbers were stripped from the file.
IMAGE_FILE_LOCAL_SYMS_STRIPPED     = 0x0008,   // COFF symbol table entries were stripped from file.
IMAGE_FILE_AGGRESIVE_WS_TRIM       = 0x0010,   // Aggressively trim the working set. This value is obsolete.
IMAGE_FILE_LARGE_ADDRESS_AWARE     = 0x0020,   // The application can handle addresses larger than 2 GB.
IMAGE_FILE_BYTES_REVERSED_LO       = 0x0080,   // The bytes of the word are reversed. This flag is obsolete.
IMAGE_FILE_32BIT_MACHINE           = 0x0100,   // The computer supports 32-bit words.
IMAGE_FILE_DEBUG_STRIPPED          = 0x0200,   // Debugging information was removed and stored separately in another file.
IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP = 0x0400,   // If the image is on removable media, copy it to and run it from the swap file.
IMAGE_FILE_NET_RUN_FROM_SWAP       = 0x0800,   // If the image is on the network, copy it to and run it from the swap file.
IMAGE_FILE_SYSTEM                  = 0x1000,   // The image is a system file.
IMAGE_FILE_DLL                     = 0x2000,   // The image is a DLL file. While it is an executable file, it cannot be run directly.
IMAGE_FILE_UP_SYSTEM_ONLY          = 0x4000,   // The file should be run only on a uniprocessor computer.
IMAGE_FILE_BYTES_REVERSED_HI       = 0x8000    // The bytes of the word are reversed. This flag is obsolete.


archtype.exe by Eli Fulkerson, August 2014
See http://www.elifulkerson.com for updates.

";

                Console.WriteLine(helpstring);
                return;
            }

            // The version string
            if (args[0] == "-v" || args[0] == "/v" || args[0] == "/version" || args[0] == "version" || args[0] == "--version")
            {
                string versionstring = @"
artchtype.exe v0.1

archtype.exe by Eli Fulkerson, August 2014
See http://www.elifulkerson.com for updates.
";
                Console.WriteLine(versionstring);
                return;
            }


            // Note here that we didn't actually have any command line arguments - both /? and /v exit the program
            // completely.  *if* we ever add any, we need to adjust args here to not parse them as filenames
            Console.WriteLine("");


            // Basic theme here - we're splitting args up by " ", then splitting those results into a Directory and
            // Filename part.  Then we're feeding those parts into Directory.GetFiles, which will handle wildcards
            // inside of Filename and return a list of matching files in the given directory.

            // This probably doesn't do wildcard directories.  Too bad, use unix if you want `find`.
            foreach (string a in args) {
                
                string filename = Path.GetFileName(a);
                string dirname = Path.GetDirectoryName(a);

                if (dirname.Length == 0)
                {
                    dirname = ".";
                }
                

                foreach (string f in Directory.GetFiles(dirname,filename)) {
                    try
                    {
                        parseExe(Path.Combine(dirname, f));
                    }
                    catch
                    {
                        // Since we are taking in arbitrary (and even wildcarded files) I didn't think it was worthwhile
                        // tracking down every single possible not-an-exe or file-wrong-size or whatever errors and
                        // giving them their own messages.
                        Console.WriteLine("General error, somehow we didn't like " + Path.Combine(dirname, f) + " :(");
                    }
                    Console.WriteLine();
                }
            }
            return;            
        }


        static void parseExe(string filename)
        {
            
            byte[] bytes;

            try
            {
                bytes = File.ReadAllBytes(filename);
            }

            catch
            {
                Console.WriteLine("Could not open specified file, aborting.");
                // removing some extra junk
                filename = filename.Replace(".\\", "");
                Console.WriteLine("Filename was : " + filename);
                return;
            }

            // removing some extra junk
            filename = filename.Replace(".\\", "");

            byte[] Slice = new List<byte>(bytes).GetRange(0, 2).ToArray();

            if (System.Text.Encoding.ASCII.GetString(Slice) != "MZ")
            {
                Console.WriteLine("Format error: First bytes not 'MZ', aborting.");
                Console.WriteLine("Filename was : " + filename);
                return;
            }

            Slice = new List<byte>(bytes).GetRange(60, 4).ToArray();

            int offset = BitConverter.ToInt32(Slice, 0);

            Slice = new List<byte>(bytes).GetRange(offset, 4).ToArray();

            if (bytes[offset] != 80 || bytes[offset + 1] != 69 || bytes[offset + 2] != 0 || bytes[offset + 3] != 0)
            {
                Console.WriteLine("Format error:  PE\\0\\0 not found at offset 60, aborting.");
                Console.WriteLine("Filename was : " + filename);
                return;
            }

            Console.WriteLine("        File : " + filename);

            offset += 4;
            byte[] Machine = new List<byte>(bytes).GetRange(offset, 2).ToArray();
            Array.Reverse(Machine);
            string machine_bits = ByteArrayToString(Machine);
            string machine_type = "Undetermined";

            if (machine_bits == "01-4C")
            {
                machine_type = "i386 (32-bit)";
            }

            if (machine_bits == "02-00")
            {
                machine_type = "IA64 (Itanium)";
            }

            if (machine_bits == "86-64")
            {
                machine_type = "x86_64 (64-bit)";
            }

            // Architecture
            Console.WriteLine("Architecture : " + machine_type);

            //@@ hey, might as well parse out the rest of the header.  At this point the bloat begins!
            

            offset += 2;
            byte[] NumberOfSections = new List<byte>(bytes).GetRange(offset, 2).ToArray();
            //Console.WriteLine(ByteArrayToString(NumberOfSections));

            offset += 2;
            byte[] TimeDateStamp = new List<byte>(bytes).GetRange(offset, 4).ToArray();
            //Console.WriteLine(ByteArrayToString(TimeDateStamp));

            uint compile_date = BitConverter.ToUInt32(TimeDateStamp, 0);
            Console.WriteLine("  Build Date : " + UnixTimeStampToDateTime(compile_date).ToString("u").Replace("Z", ""));


            offset += 4;
            byte[] PointerToSymbolTable = new List<byte>(bytes).GetRange(offset, 4).ToArray();
            //Console.WriteLine(ByteArrayToString(PointerToSymbolTable));

            offset += 4;
            byte[] NumberOfSymbols = new List<byte>(bytes).GetRange(offset, 4).ToArray();
            //Console.WriteLine(ByteArrayToString(NumberOfSymbols));

            uint symbol_pointer = BitConverter.ToUInt32(PointerToSymbolTable, 0);
            uint symbol_number = BitConverter.ToUInt32(NumberOfSymbols, 0);

            if (symbol_pointer == 0)
            {
                Console.WriteLine("     Symbols : None");
            }
            else 
            {
                Console.WriteLine("     Symbols : " + symbol_number);
            }



            
            offset += 4;
            byte[] SizeOfOptionalHeader = new List<byte>(bytes).GetRange(offset, 2).ToArray();
            //Console.WriteLine(ByteArrayToString(SizeOfOptionalHeader));

            offset += 2;
            byte[] Characteristics = new List<byte>(bytes).GetRange(offset, 2).ToArray();
            //Console.WriteLine(ByteArrayToString(Characteristics));

            //
            // I fully realize this is an asinine way to a bitfield, but I just want to get it done rather
            // than solving it the C# way at the moment...
            //

            string bits = Convert.ToString(Characteristics[0], 2).PadLeft(8, '0');
            string bits2 = Convert.ToString(Characteristics[1], 2).PadLeft(8, '0');

            string spacer = "               ";
            Console.Write("       Flags : ");

            if (bits[7] == '1')
            {
                Console.WriteLine("IMAGE_FILE_RELOCS_STRIPPED");
                Console.Write(spacer);
            }
            if (bits[6] == '1')
            {
                Console.WriteLine("IMAGE_FILE_EXECUTABLE_IMAGE");
                Console.Write(spacer);
            }
            if (bits[5] == '1')
            {
                Console.WriteLine("IMAGE_FILE_LINE_NUMS_STRIPPED");
                Console.Write(spacer);
            }
            if (bits[4] == '1')
            {
                Console.WriteLine("IMAGE_FILE_LOCAL_SYMS_STRIPPED");
                Console.Write(spacer);
            }
            if (bits[3] == '1')
            {
                Console.WriteLine("IMAGE_FILE_AGGRESIVE_WS_TRIM");
                Console.Write(spacer);
            }
            if (bits[2] == '1')
            {
                Console.WriteLine("IMAGE_FILE_LARGE_ADDRESS_AWARE");
                Console.Write(spacer);
            }
            if (bits[1] == '1')
            {
                Console.WriteLine("UNDEFINED (0x0040)");
                Console.Write(spacer);
            }

            if (bits[0] == '1')
            {
                Console.WriteLine("IMAGE_FILE_BYTES_REVERSED_LO");
                Console.Write(spacer);
            }

            if (bits2[7] == '1')
            {
                Console.WriteLine("IMAGE_FILE_32BIT_MACHINE");
                Console.Write(spacer);
            }

            if (bits2[6] == '1')
            {
                Console.WriteLine("IMAGE_FILE_DEBUG_STRIPPED");
                Console.Write(spacer);
            }

            if (bits2[5] == '1')
            {
                Console.WriteLine("IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP");
                Console.Write(spacer);
            }

            if (bits2[4] == '1')
            {
                Console.WriteLine("IMAGE_FILE_NET_RUN_FROM_SWAP");
                Console.Write(spacer);
            }

            if (bits2[3] == '1')
            {
                Console.WriteLine("IMAGE_FILE_SYSTEM");
                Console.Write(spacer);
            }

            if (bits2[2] == '1')
            {
                Console.WriteLine("IMAGE_FILE_DLL");
                Console.Write(spacer);
            }

            if (bits2[1] == '1')
            {
                Console.WriteLine("IMAGE_FILE_UP_SYSTEM_ONLY");
                Console.Write(spacer);
            }

            if (bits2[0] == '1')
            {
                Console.WriteLine("IMAGE_FILE_BYTES_REVERSED_HI");
                Console.Write(spacer);
            }

            if (bits == "00000000" && bits2 == "00000000") {
                Console.WriteLine("none");
            }

            return;
        }

        public static string ByteArrayToString(byte[] ba)
        {
            //if (BitConverter.IsLittleEndian) { Array.Reverse(ba); }
            return BitConverter.ToString(ba);
            
        }

        public static DateTime UnixTimeStampToDateTime(double unixTimeStamp)
        {
            // Unix timestamp is seconds past epoch
            System.DateTime dtDateTime = new DateTime(1970, 1, 1, 0, 0, 0, 0, System.DateTimeKind.Utc);
            dtDateTime = dtDateTime.AddSeconds(unixTimeStamp).ToLocalTime();
            return dtDateTime;
        }
    }
}
